/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.eval;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;
import java.lang.reflect.*;

// このクラスはスレッドセーフではありません。
// 複数の文インタプリタが同時に走る事は出来ますが、
// 一つの文インタプリタ内で同時にシステムコールを呼ぶ事は出来ません。

public class SystemCall extends Function
{
	Method function;
	Hashtable cache; // {関数名 => SystemCall}
	
	public SystemCall(Aya aya) {
		// ファクトリを作るコンストラクタ。
		super(aya,null,null);
		this.function = null;
		this.cache = new Hashtable();
	}
	
	protected SystemCall(Aya aya,String name,Method function) {
		super(aya,name,null);
		this.function = function;
		this.cache = null;
	}
	
	public SystemCall get(String name) {
		// 該当するシステムコールが無ければnullを返す。
		SystemCall cached = (SystemCall)cache.get(name);
		if (cached != null) {
			return cached;
		}
		
		try {
			Method function = this.getClass().getMethod("syscall_"+name,new Class[] {Vector.class});
			SystemCall syscall = new SystemCall(getAya(),name,function);
			cache.put(name,syscall);
			return syscall;
		}
		catch (Exception e) {
			return null;
		}
	}
	
	public Value eval(Vector args) throws Exception {
		// args: 要素がVariablePointerまたはValueである配列。nullでも良い。
		try {
			return (Value)function.invoke(this,new Object[] {args});
		}
		catch (Exception e) {
			throw new Exception("Calling systemcall failured: "+getName()+"("+(args == null ? "" : args.toString())+")",e);
		}
	}
	
	// ---------------- 汎用 ---------------- //
	
	public Value syscall_LOGGING(Vector args) throws Exception {
		// 本来は第一引数の式とその値をファイルに出力する関数だが、
		// プリプロセッサでも使わない限り実装が面倒なので、やるとしても後で。
		System.err.println("LOGGING: "+content(args.elementAt(0)).getString());
		return null;
	}
	
	public Value syscall_RAND(Vector args) throws Exception {
		// RAND(n)
		// 0〜n-1 の範囲でランダム値を得ます。	
		// n は省略可能で、この場合0〜99の乱数を生成します。
		long limit;
		if (args != null && args.size() > 0) {
			limit = content(args.elementAt(0)).getInteger();
		}
		else {
			limit = 100;
		}
		
		if (limit == 0) {
			return new Value(0);
		}
		else {
			return new Value((long)(Math.random() * Long.MAX_VALUE) % limit);
		}
	}
	
	public Value syscall_ASC(Vector args) throws Exception {
		// ASC(c)
		// ASCII文字を返します。
		// 指定範囲は 0≦c＜0x80 で、範囲外の c を与えると半角空白(20H)に代替されます。
		long n = content(args.elementAt(0)).getInteger();
		if (n < 0 || n >= 0x80) {
			n = 0x20;
		}
		
		char c = (char)n;
		return new Value(""+c);
	}
	
	public Value syscall_IASC(Vector args) throws Exception {
		// IASC("foo")
		// 文字列先頭の 1 バイトの文字コードを返します。半角文字しか処理できないことに注意してください。
		// 数値や空文字列を与えると -1 が返ります。
		byte[] bytes = content(args.elementAt(0)).getString().getBytes(Aya.CHARSET);
		return new Value(bytes[0]);
	}
	
	public Value syscall_ISINSIDE(Vector args) throws Exception {
		// ISINSIDE(val,-5,10)
		// ISINSIDE(val,"foo","bar")
		// 数値および文字列が指定範囲内にあるかチェックします。
		// 引数は 3 つです。第一引数はチェック対象の値(A)。第二、三引数は範囲を指定する値(i0、i1）です。
		// i0 ≦ A ≦ i1 で 1 、それ以外で 0 が返ります。i0 と i1 の大小は気にしなくて構いません。i0 のほうが大きくても正しく動作します。
		// チェック対象の価が文字列の場合は辞書順で比較されます。
		Value attention = content(args.elementAt(0));
		if (attention.isNumeric()) {
			double n = attention.getReal();
			double x = content(args.elementAt(1)).getReal();
			double y = content(args.elementAt(2)).getReal();
			
			return new Value((x <= n && n <= y) || (y <= n && n <= x) ? 1 : 0);
		}
		else {
			String str = attention.getString();
			String x = content(args.elementAt(1)).getString();
			String y = content(args.elementAt(2)).getString();
			
			return new Value((x.compareTo(str) <= 0 && str.compareTo(y) <= 0) ||
							 (y.compareTo(str) <= 0 && str.compareTo(x) <= 0) ? 1 : 0);
		}
	}
	
	// ---------------- 型チェック ---------------- //
	
	public Value syscall_ISREAL(Vector args) throws Exception {
		// ISREAL(n)
		// 値が数値（実数）かチェックします。実数には当然ながら整数が含まれます。
		// 戻り値　0/1 = 文字列/実数。
		return new Value(content(args.elementAt(0)).isNumeric() ? 1 : 0);
	}
	
	public Value syscall_ISINTEGER(Vector args) throws Exception {
		// ISINTEGER(n)
		// 値が（小数点以下の位がない）整数かチェックします。
		// 戻り値　0/1 = 文字列 or 小数点以下の位がある数値/整数。 
		return new Value(content(args.elementAt(0)).isInteger() ? 1 : 0);
	}
	
	public Value syscall_ISFUNCTION(Vector args) throws Exception {
		// ISFUNCTION("foo")
		// 文字列が関数名かチェックします。
		// 戻り値　0/1/2 = 関数名ではない/関数名/システム関数名。
		String name = content(args.elementAt(0)).getString();
		
		Function f = getAya().getDictionary().getFunction(name);
		if (f == null) {
			return new Value(0);
		}
		else if (f instanceof SystemCall) {
			return new Value(2);
		}
		else if (f instanceof Function) {
			return new Value(1);
		}
		else {
			throw new Exception();
		}
	}
	
	// ---------------- 型変換 ---------------- //
	
	public Value syscall_TONUMBER(Vector args) throws Exception {
		// TONUMBER(var)
		// システム関数 TONUMBER は、変数に格納された文字列を数値に変換します。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.store(vptr.fetch().setReal());
		return null;
	}
	
	public Value syscall_TOSTRING(Vector args) throws Exception {
		// TOSTRING(var)
		// TOSTRING は TONUMBER と逆の操作、すなわち数値から文字列への変換を行います。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.store(vptr.fetch().setString());
		return null;
	}
	
	public Value syscall_TONUMBER2(Vector args) throws Exception {
		// TONUMBER2("25")
		// TONUMBER2 は TONUMBER と機能は同等ですが、結果は戻り値として出力し、引数の内容にはタッチしません。
		return ((Value)content(args.elementAt(0)).clone()).setReal();
	}
	
	public Value syscall_TOSTRING2(Vector args) throws Exception {
		// TOSTRING2(12)
		// TOSTRING2 は TOSTRING と機能は同等ですが、結果は戻り値として出力し、引数の内容にはタッチしません。
		return ((Value)content(args.elementAt(0)).clone()).setString();
	}
	
	public Value syscall_TOBINSTR(Vector args) throws Exception {
		// TOBINSTR(3)
		// TOBINSTR は 数値を 2 進数表記の文字列へ変換します。
		return new Value(Long.toBinaryString(content(args.elementAt(0)).getInteger()));
	}
	
	public Value syscall_TOHEXSTR(Vector args) throws Exception {
		// TOHEXSTR(5)
		// TOBINSTR は 数値を 16 進数表記の文字列へ変換します。
		return new Value(Long.toHexString(content(args.elementAt(0)).getInteger()));
	}
	
	public Value syscall_BINSTRTONUM(Vector args) throws Exception {
		// BINSTRTONUM("10110")
		// BINSTRTONUM は TOBINSTR の逆操作を行います。すなわち、2 進数表記の文字列を数値へ変換します。
		return new Value(Long.parseLong(content(args.elementAt(0)).getString(),2));
	}
	
	public Value syscall_HEXSTRTONUM(Vector args) throws Exception {
		// HEXSTRTONUM("2f3a")
		// HEXSTRTONUM は TOHEXSTR の逆操作を行います。すなわち、16 進数表記の文字列を数値へ変換します。
		return new Value(Long.parseLong(content(args.elementAt(0)).getString(),16));
	}
	
	// ---------------- 簡易配列関連 ---------------- //
	
	public Value syscall_ARRAYSIZE(Vector args) throws Exception {
		// ARRAYSIZE("foo,bar,baz")
		// ARRAYSIZE("foo,bar,baz",num)
		// ARRAYSIZE システム関数は、簡易配列の要素数を求めます。
		// 引数は 1 つもしくは 2 つです。第一引数は要素数を調べる文字列。第二引数は結果を取得する変数ですが、これを省略すると結果を戻り値から取得できます。
		// この関数に数値を渡してはいけません。渡した場合 0 が返ります。
		Value result = null;
		
		Object obj = args.elementAt(0);
		if (obj instanceof Value) {
			Value val = (Value)obj;
			if (val.isNumeric()) {
				result = new Value(0);
			}
			else {
				// デフォルトはカンマ区切り
				String str = val.getString();
				int n_comma = 0;
				for (int i = 0; i < str.length(); i++) {
					if (str.charAt(i) == ',') {
						n_comma++;
					}
				}
				result = new Value(n_comma + 1);
			}
		}
		else if (obj instanceof VariablePointer) {
			VariablePointer vptr = (VariablePointer)obj;
			if (vptr.getVariableReference().getIndex() != null) {
				throw new Exception("ARRAYSIZE to element of array.");
			}
			result = new Value(vptr.getVariable().getArraySize());
		}
		else {
			throw new Exception();
		}
		
		if (args.size() >= 2) {
			VariablePointer vptr = (VariablePointer)args.elementAt(1);
			vptr.store(result);
		}
		return result;
	}
	
	public Value syscall_SETSEPARATOR(Vector args) {
		// SETSEPARATOR(foo,"-")
		// 簡易配列のセパレータは標準では半角カンマ "," ですが、SETSEPATRATOR システム関数を使用することによって、変数ごとに異なる文字列へ変更することができます。
		// ただし、システム変数のセパレータはデフォルトの半角カンマから変更することはできません。
		// 
		// 引数は 2 つで、第一引数に新しいセパレータを設定する変数、第二引数に新しいセパレータ文字列を指定します。戻り値はありません。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.getVariable().setSeparator(content(args.elementAt(1)).getString());
		return null;
	}
	
	public Value content(Object obj) {
		// objがValueならそのまま、VariablePointerならデリファレンスして返す。
		if (obj instanceof Value) {
			return (Value)obj;
		}
		else if (obj instanceof VariablePointer) {
			return ((VariablePointer)obj).fetch();
		}
		else {
			throw new RuntimeException("Internal Error: an object of "+obj.getClass().getName()+" was in arguments of system call illegally.");
		}
	}
}
